window.Gleam.initSearch([{"doc":"dijkstra","title":"dijkstra","content":"# dijkstra\n\n[![Package Version](https://img.shields.io/hexpm/v/dijkstra)](https://hex.pm/packages/dijkstra)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/dijkstra/)\n\nA versatile implementation of Dijkstra&#39;s shortest path algorithm.\n\nKey features:\n\n- No reliance on any particular graph representation.\n- Requires only a function that returns a node&#39;s successors.\n- Supports lazy graph descriptions, since nodes are only explored as required.\n- Only a single dependency: `gleamy_structures` for its fine priority queue implementation.\n\n\nFurther documentation can be found at &lt;https://hexdocs.pm/dijkstra&gt;.\n\n\n## Installation\n\n```sh\ngleam add dijkstra\n```\n\n\n## Example Usage\n\n```gleam\nimport dijkstra\n\npub fn main() {\n  let successor_func = fn(node_id: NodeId) -&gt; Dict(NodeId, Int) {\n    case node_id {\n      0 -&gt; dict.from_list([#(1,4), #(2,3)])\n      1 -&gt; dict.from_list([#(3,5)])\n      2 -&gt; dict.from_list([#(3,5)])\n      3 -&gt; dict.from_list([])\n      _ -&gt; panic as &quot;unreachable&quot;\n    }\n  }\n  \n  let shortest_paths = dijkstra.dijkstra(successor_func, 0)\n  let shortest_path_to_3 = dijkstra.shortest_path(shortest_paths, 3)\n  io.print(&quot;Shortest path to node 3 has length: &quot;)\n  io.debug(shortest_path_to_3.1)\n  io.println(&quot;.&quot;)\n  io.print(&quot;That path is: &quot;)\n  io.debug(shortest_path_to_3.0)\n  io.println(&quot;.&quot;)\n}\n```\n\n\n## Development\n\n```sh\ngleam test  # Run the tests\n```\n\n\n## Implementation Notes\n\nInspired by https://ummels.de/2014/06/08/dijkstra-in-clojure/\n\nInstead of relying on a graph data structure, node neighbours are provided by a function. Not only does this provide easy integration with any graph representation, it also works without a graph. Nodes can be generated on demand.\n\nThis is similar in concept to laziness, where elements are only generated as they&#39;re required. Not only does this open the door to use with huge graphs without having to pre-calculate unreachable regions, it also permits dynamically generated graphs.\n","url":"index.html"},{"doc":"dijkstra","title":"AllShortestPaths","content":"pub type AllShortestPaths(node_id) {\n  AllShortestPaths(\n    distances: Dict(node_id, Int),\n    predecessors: Dict(node_id, List(node_id)),\n  )\n}\n Same as [`ShortestPaths`](#ShortestPaths), except for [`dijkstra_all`](#dijkstra_all).\n The `distances` field is the same as `ShortestPaths` since there is only one shortest\n distance. But the `predecessors` field has a list of predecessors for each node instead\n of a single predecessor, to represent the possibility of there being multiple paths\n that result in the same shortest distance.\n\nAllShortestPaths(\n  distances: Dict(node_id, Int),\n  predecessors: Dict(node_id, List(node_id)),\n)\n\n\nSynonyms:\ndijkstra.AllShortestPaths\ndijkstra AllShortestPaths","url":"dijkstra.html#AllShortestPaths"},{"doc":"dijkstra","title":"ShortestPaths","content":"pub type ShortestPaths(node_id) {\n  ShortestPaths(\n    distances: Dict(node_id, Int),\n    predecessors: Dict(node_id, node_id),\n  )\n}\n The return type of [`dijkstra`](#dijkstra). Consists of two dictionaries that contain,\n for every visited node, the shortest distance to that node and the node&#39;s immediate\n predecssor on that shortest path.\n\nShortestPaths(\n  distances: Dict(node_id, Int),\n  predecessors: Dict(node_id, node_id),\n)\n\n\nSynonyms:\ndijkstra.ShortestPaths\ndijkstra ShortestPaths","url":"dijkstra.html#ShortestPaths"},{"doc":"dijkstra","title":"SuccessorsFunc","content":"pub type SuccessorsFunc(node_id) =\n  fn(node_id) -&gt; Dict(node_id, Int)\n A function that given a node, returns successor nodes and their distances.\n\n\nSynonyms:\ndijkstra.SuccessorsFunc\ndijkstra SuccessorsFunc","url":"dijkstra.html#SuccessorsFunc"},{"doc":"dijkstra","title":"dijkstra","content":"pub fn dijkstra(\n  edges_from: fn(a) -&gt; Dict(a, Int),\n  start: a,\n) -&gt; ShortestPaths(a)\n Run Dijkstra&#39;s algorithm to determine the shortest path to every node reachable from\n `start`, according to `edges_from`.\n\n ## Examples\n\n \n let f = fn(node_id: NodeId) -&gt; Dict(NodeId, Int) {\n   case node_id {\n     0 -&gt; dict.from_list([#(1,4), #(2,3)])\n     1 -&gt; dict.from_list([#(3,5)])\n     2 -&gt; dict.from_list([#(3,5)])\n     3 -&gt; dict.from_list([])\n     _ -&gt; panic as &quot;unreachable&quot;\n   }\n }\n\n dijkstra.dijkstra(f, 0)\n // -&gt; ShortestPaths(dict.from_list([#(0, 0), #(1, 4), #(2, 3), #(3, 8)]), dict.from_list([#(1, 0), #(2, 0), #(3, 2)]))\n \n\nSynonyms:\ndijkstra.dijkstra\ndijkstra dijkstra","url":"dijkstra.html#dijkstra"},{"doc":"dijkstra","title":"dijkstra_all","content":"pub fn dijkstra_all(\n  edges_from: fn(a) -&gt; Dict(a, Int),\n  start: a,\n) -&gt; AllShortestPaths(a)\n Same as [`dijkstra`](#dijkstra), except each node predecessor is a `list` instead of a\n single node. If there are multiple shortest paths, junction nodes will have more than one\n predecessor.\n \nSynonyms:\ndijkstra.dijkstra_all\ndijkstra dijkstra_all","url":"dijkstra.html#dijkstra_all"},{"doc":"dijkstra","title":"has_path_to","content":"pub fn has_path_to(paths: ShortestPaths(a), dest: a) -&gt; Bool\n Return true if Dijkstra&#39;s algorithm found a path to the `dest` node.\n \n Recall that in order to determine the shortest path, Dijkstra&#39;s algorithm visits all\n nodes reachable from the given start node. Thus we can exploit that to determine whether\n any particular node is reachable, without looking at the graph again.\n\nSynonyms:\ndijkstra.has_path_to\ndijkstra has_path_to","url":"dijkstra.html#has_path_to"},{"doc":"dijkstra","title":"shortest_path","content":"pub fn shortest_path(\n  paths: ShortestPaths(a),\n  dest: a,\n) -&gt; #(List(a), Int)\n When applied to the result of [`dijkstra`](#dijkstra), returns the shortest path to the\n `dest` node as a list of successive nodes, and the total length of that path.\n\nSynonyms:\ndijkstra.shortest_path\ndijkstra shortest_path","url":"dijkstra.html#shortest_path"},{"doc":"dijkstra","title":"shortest_paths","content":"pub fn shortest_paths(\n  all_paths: AllShortestPaths(a),\n  dest: a,\n) -&gt; #(List(List(a)), Int)\n Same as [`shortest_path`](#shortest_path), except for [`dijkstra_all`](#dijkstra_all).\n \nSynonyms:\ndijkstra.shortest_paths\ndijkstra shortest_paths","url":"dijkstra.html#shortest_paths"},{"doc":"dijkstra","title":"dijkstra","content":"\n The key to this library is the user specified [`SuccessorsFunc`](#SuccessorsFunc) function,\n passed as the `edges_from` parameter, with type `fn(node_id) -&gt; Dict(node_id, Int)`.\n \n To define a suitable function, first define a type for the `node_id` type variable. For many\n graphs a simple `Int` or `String` will suffice to identify each node. But you are free to\n include context in the type, such as direction or a counter, by specifying a tuple or record.\n You can then use this context to determine the successor nodes.\n \n `SuccessorsFunc` takes a `node_id` of the type you define, and returns a `Dict` of successor\n nodes and their distances. In traditional graph theory, these are called *edges*. For\n example, a simple undirected graph could be represented with a `node_id` of `Int` and a\n return value consisting of all the adjacent nodes and their distances. But you could just\n as well represent a game of tic-tac-toe, where each `node_id` is the current game state and\n each successor is a possible next move.\n \n If you already have a graph type, it&#39;s likely you need only map `SuccessorsFunc` to that\n graph type&#39;s function for listing outgoing nodes. For example, if you&#39;re using the\n [`graph`](https://hex.pm/packages/graph) package, then your `SuccessorsFunc` might look\n like this:\n \n ```gleam\n type NodeId = Int\n let f: SuccessorsFunc(NodeId) = fn(node: NodeId) {\n   let my_graph =\n     graph.new()\n     |&gt; graph.insert_node(1, &quot;one node&quot;)\n     |&gt; graph.insert_node(2, &quot;other node&quot;)\n     |&gt; graph.insert_directed_edge(&quot;edge label&quot;, from: 1, to: 2)\n     |&gt; ...\n   \n   let assert Ok(context) = graph.get_context(my_graph, node)\n   context.outgoing\n   |&gt; dict.keys\n   |&gt; list.map(fn(id) { #(id, 1) })\n   |&gt; dict.from_list\n }\n \n For graphs without edge weights, it is perfectly valid to use the same distance value for\n every successor.\n Typically however, you don&#39;t want to create the graph inside your `SuccessorsFunc` function.\n So note that the function can capture immutable data and just use it to determine the\n successors. For example you could have a function that accepts a custom data store, that\n creates your `SuccessorsFunc` function by capturing that store and querying it based on the\n node provided. Suppose you were trying to make a certain paragraph from a list of sentence\n fragments. You could define the node as the paragraph left to construct, capture the list\n of sentence fragments in your `SuccessorsFunc`, and check if any of them can help complete\n your paragraph. It might look something like this:\n \n ```gleam\n type NodeId = String\n fn make_successor_fn(snippets: List(String)) -&gt; SuccessorsFunc(NodeId)\n {\n   fn(remaining_text: NodeId) -&gt; Dict(NodeId, Int) {\n     list.filter_map(remaining_text, fn (snippet) {\n       case string.starts_with(remaining_text, snippet) {\n         False -&gt; Error(Nil)\n         True  -&gt; Ok(#(string.drop_start(remaining_text, string.length(snippet)), 1))\n       }\n     })\n     |&gt; dict.from_list\n   }\n }\n \n Then you simply pass your data store to the generator to get your function:\n \n ```gleam\n let can_be_done = make_graph_fn(my_snippets)\n |&gt; dijkstra.dijkstra(my_paragrah)\n |&gt; dijkstra.has_path_to(&quot;&quot;)\n ```\n \n You can even treat your `node_id` as a position with some state. So for example if you had\n a maze represented as a matrix of integer coordinates, but you want the ability to walk\n through walls once on your way to the exit, you could define your `node_id` as\n `#(#(Int, Int), Int)`. The inner pair is the current coordinates, and the last `Int`\n is `1` if you have a wall-walk left or `0` if you&#39;ve used it up. Your `SuccessorsFunc`\n generator might then look something like this:\n \n ```gleam\n type Coord = #(Int, Int)\n type NodeId = #(Coord, Int)\n fn make_successor_fn(m: Matrix) -&gt; SuccessorsFunc(NodeId)\n {\n   fn(n: NodeId) -&gt; Dict(NodeId, Int) {\n     let #(coord, cheats_left) = n\n     list.filter_map(get_neighbours(m, coord), fn(neighbour) {\n       case is_wall(m, neighbour), cheats_left &gt; 0 {\n         True,  False -&gt; Error(Nil)                         //Wall and no cheats left\n         True,  True  -&gt; Ok(#(neighbour, cheats_left - 1))  //Wall and can use a cheat\n         False, _     -&gt; Ok(#(neighbour, cheats_left))      //No wall, free to proceed\n           }\n       }\n     })\n     |&gt; list.map(fn(coord_cheats_left) {\n       #(coord_cheats_left, 1) //all moves cost the same\n     })\n     |&gt; dict.from_list\n   }\n }\n\n Finally, instead of deciding next moves based on the current state, your `SuccessorsFunc`\n could determine a distance based on the current state. Say for example you&#39;re modelling\n a powerboat that is fastest in a straight line. Your `node_id` could include the direction\n from which the current node was reached. Then your `SuccessorsFunc` could use that to\n prevent turning around on the spot, and to set the edge distance based on whether the\n direction is changing or not. Something like this:\n \n type NodeId = #(Coord, Direction)\n fn make_successor_fn(m: Matrix) -&gt; SuccessorsFunc(NodeId)\n {\n   fn(n: NodeId) -&gt; Dict(NodeId, Int) {\n     let #(coord, old_dir) = n\n     [N, E, S, W]\n     |&gt; list.filter(fn(new_dir) { new_dir != opposite(old_dir) }) //can&#39;t go backwards\n     |&gt; list.map(fn(new_dir) { #(get_neighbours(m, coord), new_dir) })\n     |&gt; list.map(fn(coord_dir) {\n       #(coord_dir, case coord_dir.1 == old_dir { True -&gt; 1  False -&gt; 10 })\n     })\n     |&gt; dict.from_list\n   }\n }\n \n So you can see that decoupling this library from a particular graph implementation makes\n it applicable to a wide variety of problems. By designing a suitable `SuccessorsFunc`\n function, and taking advantage of function captures and the variable `node_id` type, you\n can apply Dijkstra&#39;s algorithm to scenarios that would otherwise be difficult to model as\n a graph. In my experience, the most difficult part of applying Dijkstra from a library is\n often working out how to transform your problem domain to the graph representation expected\n of the library. Eliminating that dependency might mean more time exploring the problem\n domain and less time learning a new graph abstraction.","url":"dijkstra.html"}]);